package com.danitheskunk.skunkworks.gfx.vt;

import com.danitheskunk.skunkworks.Vec2i;
import com.danitheskunk.skunkworks.gfx.Color;
import com.danitheskunk.skunkworks.gfx.font.IFont;

import java.io.UnsupportedEncodingException;
import java.nio.ByteBuffer;
import java.nio.IntBuffer;
import java.util.ArrayList;
import java.util.List;

public class Terminal {
	private List<Cell> cells;
	private Vec2i size;
	private Vec2i fullCharSize;
	private Vec2i halfCharSize;
	private IFont fullFont;
	private IFont halfFont;

	public Terminal(Vec2i size, IFont fullFont, IFont halfFont) {
		this.cells = new ArrayList<>();
		this.size = size;
		this.fullFont = fullFont;
		this.halfFont = halfFont;

		if(!fullFont.isMonospace() || !halfFont.isMonospace()) {
			throw new RuntimeException("Fonts need to be monospace");
		}

		this.fullCharSize = fullFont.getMonospaceSize();
		this.halfCharSize = halfFont.getMonospaceSize();

		if(fullCharSize.getY() != halfCharSize.getY() || fullCharSize.getX() != halfCharSize.getX() * 2) {
			throw new RuntimeException("halfFont needs to be half width of fullFont");
		}

		for(int i = 0; i < size.getY() * size.getX(); ++i) {
			var cell = new Cell();
			cell.fullChar = 0;
			cell.secondChar = 0;
			cell.halfWidth = false;
			cell.bgColor = Color.BLACK;
			cell.fgColor = Color.WHITE;
			cells.add(cell);
		}
	}

	public Vec2i getFullCharSize() {
		return fullCharSize;
	}

	public Vec2i getHalfCharSize() {
		return halfCharSize;
	}

	public Vec2i getSize() {
		return size;
	}

	public void setChar(Vec2i pos, int ch) {
		var cell = cells.get(vecToPos(pos));
		cell.fullChar = ch;
		cell.halfWidth = false;
	}

	public void setChar(Vec2i pos, int ch, Color foregroundColor) {
		var cell = cells.get(vecToPos(pos));
		cell.fullChar = ch;
		cell.halfWidth = false;
		cell.fgColor = foregroundColor;
	}

	public void setChar(Vec2i pos, int ch, Color foregroundColor, Color backgroundColor) {
		var cell = cells.get(vecToPos(pos));
		cell.fullChar = ch;
		cell.halfWidth = false;
		cell.fgColor = foregroundColor;
		cell.bgColor = backgroundColor;
	}

	public void setHalfChars(Vec2i pos, int chLeft, int chRight) {
		var cell = cells.get(vecToPos(pos));
		cell.fullChar = chLeft;
		cell.secondChar = chRight;
		cell.halfWidth = true;
	}

	public void setHalfChars(Vec2i pos, int chLeft, int chRight, Color foregroundColor) {
		var cell = cells.get(vecToPos(pos));
		cell.fullChar = chLeft;
		cell.secondChar = chRight;
		cell.halfWidth = true;
		cell.fgColor = foregroundColor;
	}

	public void setHalfChars(Vec2i pos, int chLeft, int chRight, Color foregroundColor, Color backgroundColor) {
		var cell = cells.get(vecToPos(pos));
		cell.fullChar = chLeft;
		cell.secondChar = chRight;
		cell.halfWidth = true;
		cell.fgColor = foregroundColor;
		cell.bgColor = backgroundColor;
	}

	public void setForegroundColor(Vec2i pos, Color color) {
		var cell = cells.get(vecToPos(pos));
		cell.fgColor = color;
	}

	public void setBackgroundColor(Vec2i pos, Color color) {
		var cell = cells.get(vecToPos(pos));
		cell.fgColor = color;
	}

	public int getChar(Vec2i pos) {
		return cells.get(vecToPos(pos)).fullChar;
	}

	public int getLeftHalfChar(Vec2i pos) {
		return cells.get(vecToPos(pos)).fullChar;
	}

	public int getRightHalfChar(Vec2i pos) {
		return cells.get(vecToPos(pos)).secondChar;
	}

	public Color getForegroundColor(Vec2i pos) {
		return cells.get(vecToPos(pos)).fgColor;
	}

	public Color getBackgroundColor(Vec2i pos) {
		return cells.get(vecToPos(pos)).bgColor;
	}

	public boolean isHalfWidth(Vec2i pos) {
		return cells.get(vecToPos(pos)).halfWidth;
	}

	private int vecToPos(Vec2i pos) {
		return pos.getX() + pos.getY() * size.getX();
	}

	public IFont getFullFont() {
		return fullFont;
	}

	public IFont getHalfFont() {
		return halfFont;
	}

	public Vec2i getPixelSize() {
		return Vec2i.mul(size, fullCharSize);
	}

	public void drawString(Vec2i pos, String str) {
		drawString(pos, str, null);
	}

	public void drawString(Vec2i pos, String str, Color foregroundColor) {
		drawString(pos, str, foregroundColor, null);
	}

	public void drawString(Vec2i pos, String str, Color foregroundColor, Color backgroundColor) {
		IntBuffer buf;

		try {
			buf = ByteBuffer.wrap(str.getBytes("UTF-32")).asIntBuffer();
		} catch(UnsupportedEncodingException e) {
			throw new RuntimeException(e);
		}

		for(int i = 0; i < str.length(); ++i) {
			var ch = buf.get(i);
			var cpos = new Vec2i(pos.getX() + i, pos.getY());
			if(ch > 0 && ch < 256) {
				setChar(cpos, ch);
			}
			if(foregroundColor != null) {
				setForegroundColor(cpos, foregroundColor);
			}
			if(backgroundColor != null) {
				setBackgroundColor(cpos, backgroundColor);
			}
		}
	}

	public void drawHalfString(Vec2i pos, String str) {
		drawHalfString(pos, str, null);
	}

	public void drawHalfString(Vec2i pos, String str, Color foregroundColor) {
		drawHalfString(pos, str, foregroundColor, null);
	}

	public void drawHalfString(Vec2i pos, String str, Color foregroundColor, Color backgroundColor) {
		IntBuffer buf;

		try {
			buf = ByteBuffer.wrap(str.getBytes("UTF-32")).asIntBuffer();
		} catch(UnsupportedEncodingException e) {
			throw new RuntimeException(e);
		}

		for(int i = 0; i < str.length(); i += 2) {
			var ch1 = buf.get(i);
			var ch2 = i + 1 == str.length() ? 0 : buf.get(i + 1);
			var cpos = new Vec2i(pos.getX() + i / 2, pos.getY());

			if(ch1 < 0 || ch1 >= 256) ch1 = 0;
			if(ch2 < 0 || ch2 >= 256) ch2 = 0;

			setHalfChars(cpos, ch1, ch2);

			if(foregroundColor != null) {
				setForegroundColor(cpos, foregroundColor);
			}
			if(backgroundColor != null) {
				setBackgroundColor(cpos, backgroundColor);
			}
		}
	}

	private class Cell {
		int fullChar;
		int secondChar;
		boolean halfWidth;
		Color bgColor;
		Color fgColor;
	}

}
